
## ADVANCED ECONOMETRICS
#
#  SIMULATE DATA FROM GARCH(1,1) MODEL
#
#  Description:
#  This code snippet shows how to simulate data from a generalized
#  autoregressive conditional hetereoeskedasticity (GARCH) model given by:
#
#  x(t) = sqrt(sig(t)) * eps(t)
#
#  sig(t+1) = omega + alpha * x(t)^2 + beta * sig(t)
#
#  where eps ~ NID(0,1)
#
#  Francisco Blasques 2019

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
## 1. Setup

T=1000  # sample size

## 2. Parameter Values

omega = 0.1      # intercept parameter in update equation
alpha = 0.05     # OD parameter in update equation
beta =  0.94     # autoregressive parameter in update equation
sig1 = omega/(1-alpha-beta) # initial value for conditional volatility

## 3. Generate Innovations

epsilon = np.random.randn(T,1)         # generate a vector of T random normal


## 4. Define Time Series Vector

x = np.zeros((T,1)); # define vector of zeros of length T
sig = np.zeros((T+1,1));
## 5. Define Initialization for Time Series
sig[0,] = sig1


## 6. Generate Time Series

for t in range(0, T):
    x[t,:] = np.sqrt(sig[t,:]) * epsilon[t,:] # generate x(t) recursively
    sig[(t+1),:] = omega + alpha * x[t,:]**2 + beta * sig[t,:]

## 7. Plot Data

plt.plot(x)
plt.plot(sig[0:-1]**(1/2))
plt.show()
