#!/usr/bin/env python3
# -*- coding: utf-8 -*-

## ADVANCED ECONOMETRICS
#
#  STANDARD ERRORS OF ML ESTIMATOR OF GAUSSIAN AR(1) PARAMETERS
#
#  Description:
#  This code snippet shows how to obtain the standard errors
#  of the maximum likelihood estimator of the parameters  of
#  a Gaussian AR(1) model.
#
#  Francisco Blasques 2019 (translated by Janneke van Brummelen)

#%% 0. Imports
import numpy as np
import pandas as pd
import scipy.optimize as opt
from llik_fun_AR1 import llik_fun_AR1
from grad import * 

#%% 1. Set-up

dfX  = pd.read_csv('output_data.csv') # in this file the time series data is labeled "x"
vX = dfX['x'].values

#%% 2. Optimization Options

options ={'eps':1e-09,  # argument convergence criteria
         'disp': True,  # display iterations
         'maxiter':500} # maximum number of iterations

#%% 3. Initial Parameter Values

alpha_ini = 0.2  # initial value for intercept
beta_ini = 0.2   # initial value for ar coefficient
sig_ini = 0.5    # initial value for innovation variance

theta_ini = np.array([alpha_ini,
                      beta_ini,
                      sig_ini
                     ])

#%% 4. Optimize Log Likelihood Criterion

#  optim input:
# (1) negative log likelihood function: llik_fun_AR1() note the minus when calculating the mean
# (2) initial parameter: theta_ini
# (3) parameter space bounds
# (4) optimization setup: options
#  Note: a number of parameter restriction could be left empty with []

#  optim output:
# (1) parameter estimates: x
# (2) negative mean log likelihood value at theta_hat: fun
# (3) exit flag indicating (no) convergence: convergence

results = opt.minimize(llik_fun_AR1, theta_ini, args=(vX), method='L-BFGS-B', 
                                  bounds=( (-10,  10), # Parameter Space Bounds
                                  (-1, 1), (0.00001, 100) )) #restrictions in parameter space

#%% 5. Print Optimization Output

print('parameter estimates:')
print(results.x)

print('log likelihood value:')
print(results.fun)

print('Successful minimization?')
print(results.success)

#%% 6. Calculate Sigma Matrix, Omega Matrix and Std Errors

iT = vX.shape[0]
iK = results.x.shape[0]

# Calculate SE under assumption of correct specification 
mH= hessian_2sided(llik_fun_AR1, results.x, vX) 
mS2= np.linalg.inv(mH)/(iT-1)
vS= np.sqrt(np.diag(mS2))

# Calculate SE without assumption of correct specification 
mA = -mH
mAi = np.linalg.inv(mA)
mG = np.zeros((iT-1,iK)).reshape(iT-1,iK)
    
for i in range(1,iT):
    vG = gradient_2sided(llik_fun_AR1, results.x, vX[i-1:i+1])
    mG[i-1,:]= np.array(vG).reshape(1,iK)
    
mB = mG.T @ mG/(iT-1)
mBi = np.linalg.inv(mB)
vSjac = np.sqrt(np.diag(mBi.T @ np.cov(-mG.T) @ mBi / (iT-1)))

# Robust/sandwich asymptotic covariance matrix
mS2rob = (-mAi) @ mB @ (-mAi)/(iT-1)
vSrob = np.sqrt(np.diag(mS2rob))

#Calculate robust score variance allowing strong misspecification (no mds score)
RSigma = np.zeros((iK,iK))
maxlag = 10

for i in range(iK):
    for j in range(iK): 
        for l in range(maxlag+1):          
            RSigma[i,j] = RSigma[i,j] + np.average(mG[maxlag:,i] * mG[maxlag-l:iT-1-l,j]) \
            - np.average(mG[maxlag:,i] ) * np.average(mG[maxlag-l:-i-1,j])

RSigma = RSigma*(iT-1)

mS2nomds = mS2 @ RSigma @ mS2
vSnomds = np.sqrt(np.diag(mS2nomds))

#%% 7. Print Standard Error Output
print('\n SE calculated under assumption of correct specification (neg inverse Hessian)')
print(vS)

print('\n SE calculated with assumption of correct specification (jacobian approach)')
print(vSjac)

print('\n SE calculated without assumption of correct specification')
print(vSrob)

print('\n SE calculated allowing for strong misspecification (no mds score)')
print(vSnomds)

